/*
 *
 * Copyright (C) 2006-2007  imaxin|software
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * */

#include <iostream>
#include <string>

#include <pcrecpp.h>

#include "config.h"
#include "Engine.h"
#include "freeling.h"
#include "freeling/traces.h"
#ifndef UTIL_H
#define UTIL_H
#include "Util.h"
#endif

#ifdef DEBUG
#define DPRINT(var)		cout << var << std::endl
#define DPRINT(var, val) 	cout << var << ": " << val << std::endl
#else
#define DPRINT(var)
#define DPRINT(var, val)
#endif

using namespace std;

Engine::Engine()
{
	_disambiguator = new Disambiguator(
								TOKENIZER_PATH,
								SPLITTER_PATH,
								QUANTIES_PATH,
								LOCUTIONS_PATH,
								SUFFIX_PATH,
								PROBABILITY_PATH,
								DICTIONARY_PATH,
								NPDATAFILE_PATH,
								PUNCTUATION_PATH,
								TAGGER_PATH,
								LANGUAGE
								);

	_hException				= new HashException(EXCEPTION_PATH, LEMMA_PATH);
	_hError					= new HashError(LEMMA_SOLUTION_PATH);
	_hError_invariable		= new HashError(LEMMA_SOLUTION_INV_PATH);
	_hError_invariableFirst = new HashError(LEMMA_SOLUTION_INV1_PATH);
}

Engine::~Engine()
{
	delete _disambiguator;
	delete _hException;
	delete _hError;
	delete _hError_invariable;
	delete _hError_invariableFirst;
}

int Engine::check(const string text)
{
	if (this->validateInput(text)) {
		string buffer = text;
		StringUtil::Trim(buffer);
		PhraseStructure ora(buffer);

		_disambiguator->run(&ora);
		DPRINT("", text);
		DPRINT("", ora.getAnalysisConcat());

		if (this->searchException(&ora))
			return 0;
		if (this->searchError(&ora, NULL, NULL) > -1)
			return 1;
	}

	return 0;
}

int Engine::spell(const string text, string *sol) {
	if (this->validateInput(text)) {
		string buffer = text;
		StringUtil::Trim(buffer);
		PhraseStructure ora(buffer);

		_disambiguator->run(&ora);

		// Hay que descarcatar los casos de falsos errores
		if (!this->searchException(&ora)) {
			string error_replace;
			string lemma_pattern_error;
			vector<string> v_sugg;
			vector<string> v_final_suggs;
			int pos_error = -1;

			if ( (pos_error = this->searchError(&ora, &lemma_pattern_error, &v_sugg)) > -1) {
				Util::TrimVector(&v_sugg);
				Solution candidate(ora);

				// Cuando la oracin contenga ms de un elemento se deber
				// de generar la concordancia entre det + nombre
				vector<string> v_sugg_tagged = this->taggerFirstWord(v_sugg);

				candidate.buildGrammarBetweenArticleAndName(pos_error,
															v_sugg,
															v_sugg_tagged,
															&v_final_suggs,
															&error_replace
															);

				this->buildXmlSolution(error_replace, pos_error, v_final_suggs, sol);
				DPRINT("error_replace", error_replace);
				DPRINT("lema error", lemma_pattern_error);
				DPRINT("pos_error", pos_error);
				DPRINT("v_sugg.size", v_sugg.size());
			}
		}
	}

	return 0;
}

int Engine::searchException(PhraseStructure *p) {
	string analysis = p->getParolesConcat();
	NGram ngrama(analysis);
	vector<string> v_lemmas_in_phrase = p->getLemmaCollectionFilteredByName();
	int len = NGRAM_MAX;

	while (len > 1) {
		vector<string> v_tags_in_ngrama = ngrama.splitString(len);
		int size = v_tags_in_ngrama.size();
		int i = 0;

		for (i = 0; i < size; i++) {
			if (_hException->searchPattern(v_tags_in_ngrama.at(i), v_lemmas_in_phrase))
				return 1;
		}
		len--;
	}

	return 0;
}

int Engine::searchError(PhraseStructure *p, string *lemma_pattern_error, vector<string> *v_sugg) {
	string analysis = p->getAnalysisConcat();
	NGram ngrama(analysis);
	int len = NGRAM_MIN;


	while (len > 0) {
		vector<string> v_analysis_in_ngrama = ngrama.splitString(len);
		int pos_error = -1;

		if ((pos_error = this->findCollectionInHash(v_analysis_in_ngrama, v_sugg)) > -1)  {
			int type = 0;
			int pos = 0;

			if (!Util::setPosErrorAndType(pos_error, &pos, &type))
				return -1;
			// Se verifican no errores del tipo "a artista"
			if (type == ERROR_TYPE_INVARIABLE)
				if (!p->isDeterminerMaleThePreviousTag(pos))
					return -1;

			if (lemma_pattern_error)
				*lemma_pattern_error = v_analysis_in_ngrama.at(pos);

			// Al buscar el patron en el hash marca el error en la
			// posicion del SPS00 cuando debera de macarla dos posiciones
			// a la derecha, coincidiendo con un sustantivo
			return (pos_error & 0x1000) ? pos_error + 2 : pos_error;
		}

		len--;
	}

	return -1;
}

int Engine::findCollectionInHash(vector<string> v_analysis_ngram, vector<string> *v_suggestion) {
	int size = v_analysis_ngram.size();
	int i = 0;

	if (size > 1) {
		for (i = 0; i < size; i++)
			if (_hError_invariableFirst->searchPattern(v_analysis_ngram.at(i), v_suggestion))
				return (0x1000 | i);
			else if (_hError->searchPattern(v_analysis_ngram.at(i), v_suggestion))
				return (0x2000 | i);
			else if (_hError_invariable->searchPattern(v_analysis_ngram.at(i), v_suggestion))
				return (0x4000 | i);
	}
	else if (size == 1) {
		if (_hError_invariableFirst->searchPattern(v_analysis_ngram.at(0), v_suggestion))
			return 0x1000;
		if (_hError->searchPattern(v_analysis_ngram.at(0), v_suggestion))
			return 0x2000;
	}

	return -1;
}

int Engine::validateInput(string input) {
	try {
		if (input.size() > 0)
			return 1;
		else
			return 0;
	}
	catch(exception &e) { return 0; }
}

int Engine::buildXmlSolution(const string error, int pos_error, vector<string> v_sugg, string *sol) {
	string buffer;
	int i;
	int size = v_sugg.size();

	if (size > 0) {
		buffer	= "<ITEM><ERROR>" + error;
		buffer += "</ERROR><SUGGESTS>";
		for (i = 0; i < size; i++)
			buffer += "<SUGGEST>" + v_sugg.at(i) + "</SUGGEST>";

		buffer += "</SUGGESTS><REPLACE>" + error;
		buffer += "</REPLACE></ITEM>";
	}
	else
		buffer = "<ITEM><ERROR></ERROR><SUGGESTS></SUGGESTS><REPLACE></REPLACE></ITEM>";

	if (sol)
		*sol = buffer;

	return size;
}

vector<string> Engine::taggerFirstWord(vector<string> v_suggs) {
	int i;
	int size = v_suggs.size();
	vector<string> v_sugg_tagged;

	for (i = 0; i < size; i++) {
		string buffer = "";
		buffer = Util::substitute(v_suggs.at(i), "_/_", "");
		buffer = Util::substitute(buffer, "_e_", "");
		buffer = Util::substitute(buffer, "_ou_", "");

		Phrase suggestion(buffer);
		this->_disambiguator->run(&suggestion);

		v_sugg_tagged.push_back(suggestion.getParole(0));
	}

	return v_sugg_tagged;
}

